from pathlib import Path
from typing import Optional

import yaml
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from .helpers import fix_absent_signer_problem, fix_multiple_sections_in_first


def get_sections(filename: Path) -> Optional[dict]:
    """The existence of the path having been previously verified, this implies
    prior processing from the `dblegacy` library which produces a yaml file
    whose contents are in list format.

    Args:
        filename (Path): The legacy file

    Returns:
        dict: The list of section data, populated from the legacy file.
    """
    try:
        with open(filename, "r") as r:
            return yaml.load(r, Loader=yaml.FullLoader)
    except ScannerError as e:
        print(f"See error {e}")
    except ParserError as e:
        print(f"See error {e}")
    return None


def set_units(folder: Path, data: dict):
    """
    If a preformatted file exists, i.e. data sourced from the old nightshade "Republic Act" repository,
    or formatted manually for use (see rawlaw repository), use the data contained in this file. The file is uniformly named as follows: `category` + `serial_number`.`yaml`, e.g. `ra11054.yaml` for Republic Act No. 11054;

    If a preformatted file is absent,  use the unformatted `units.yaml` file contained within the statute folder. The contents of this file is contained
    in the data dictionary passed.

    Args:
        folder (Path): [description]
        data (dict): [description]

    Returns:
        [type]: [description]
    """
    preformatted = folder / f"{data['category']}{folder.parts[-1]}.yaml"
    unformatted = folder / "units.yaml"

    if preformatted.exists():
        print(f"Provisions found (preformatted): {preformatted}.")
        data["units"] = get_sections(preformatted)

    elif unformatted.exists():
        print(f"Provisions found (unformatted): {unformatted}.")
        with open(unformatted, "r") as r:
            data["units"] = yaml.load(r, Loader=yaml.FullLoader)
            data = fix_absent_signer_problem(data)
            data = fix_multiple_sections_in_first(data)

    return data



from typing import Optional

ALLOWED_CATEGORIES = [
    "ra",
    "eo",
    "pd",
    "ca",
    "bp",
    "act",
    "const",
    "veto",
    "roc",
    "am",
    "bm",
    "es",
    "res",
    "roc",
]


def get_title(category: str, num: str) -> Optional[str]:
    """Combine two strings after formatting the category

    Args:
        category (str): Determines (generally) the first part of the string; see exceptions in "roc" and "const"
        num (str): The serialized identifier of the category

    Returns:
        Optional[str]:
    """
    cat = category.lower()
    if cat not in ALLOWED_CATEGORIES:
        return None

    if cat == "ra":
        return f"Republic Act No. {num.upper()}"

    elif cat == "eo":
        return f"Executive Order No. {num.upper()}"

    elif cat == "pd":
        return f"Presidential Decree No. {num.upper()}"

    elif cat == "bp":
        return f"Batas Pambansa Blg. {num.upper()}"

    elif cat == "ca":
        return f"Commonwealth Act No. {num.upper()}"

    elif cat == "act":
        return f"Act No. {num.upper()}"

    elif cat == "bm":
        return f"Bar Matter No. {num.upper()}"

    elif cat == "am":
        return f"Administrative Matter No. {num.upper()}"

    elif cat == "roc":
        return f"{num} Rules of Court"

    elif cat == "res":
        return f"Resolution of the Court En Banc dated {num}"

    elif cat == "cir":
        return f"Circular No. {num}"

    elif cat == "const":
        return f"{num} Constitution"

    elif cat == "es":
        return f"Spanish {num}"

    elif cat == "veto":
        return f"Veto Message of Republic Act No. {num}"

    return None



import re
from pathlib import Path
from typing import Iterator, Optional

from .decisions import define_data
from .statute_parts.title import ALLOWED_CATEGORIES
from .statutes import data_from_folder


def define_statute(parent: Path, child: Path, context: str) -> Optional[dict]:
    """

    Args:
        parent (Path): The parent path
        child (Path): The path to the statute
        context (str): The kind of statute

    Returns:
        Optional[dict]: [description]
    """
    # Is the folder digit-like, e.g. 209 or 209-A or 12-34-SC
    if not re.search(r"(^\d+-[A-C]$)|(^\d+$)|[a-zA-Z0-9-]+", child.name):
        print(f"Not digit-like: {child}")
        return None

    print(f"Processing: {child}")
    if data := data_from_folder(parent, context, child.name):
        return data
    else:
        print(f"Missing data: {child}")
        return None


def decode_statutes(parent: Path, context: str) -> Optional[Iterator[dict]]:
    """Given a parent directory `parent` with a subfolder `context`, generate statute-like data from entries of the subfolder, or the grandchildren of the parent directory.

    Args:
        parent (Path): The parent local directory, e.g. rawlaw
        context (str): Possible options include "ra", "eo", "pd", "ca", "bp", "act", "const"

    Returns:
        Optional[Iterator[dict]]: [description]

    Yields:
        Iterator[Optional[Iterator[dict]]]: [description]
    """
    if context not in ALLOWED_CATEGORIES:
        return None

    folder = parent / "statutes"
    if not folder.exists():
        return None

    subfolder = folder / context
    if not subfolder.exists():
        return None

    for child in subfolder.glob("*"):
        yield define_statute(folder, child, context)


def decode_decisions(loc: Path, context: str) -> Optional[Iterator[dict]]:
    """Given a parent directory "location" with a subfolder named "context",
    generate decision-like data from entries of the subfolder, or the grandchildren of the parent directory.

    Args:
        loc (Path): The parent local directory
        context (str): Either "legacy" or "sc"

    Returns:
        Optional[Iterator[dict]]: [description]

    Yields:
        Iterator[Optional[Iterator[dict]]]: [description]
    """
    if context not in ["legacy", "sc"]:
        return None

    folder = loc / "decisions" / context
    if not folder.exists():
        return None

    locations = folder.glob("*")
    for location in locations:
        yield define_data(location)



import re
from pathlib import Path
from typing import Optional

import arrow

from .statute_parts.body import set_units
from .statute_parts.meta import set_meta
from .statute_parts.title import ALLOWED_CATEGORIES, get_title


def uniform_section_label(raw: str):
    """Replace the SECTION | SEC. | Sec. format with the word Section, if applicable."""
    regex = r"""
        ^\s*
        S
        (
            ECTION|
            EC|
            ec
        )
        [\s.,]+
    """
    pattern = re.compile(regex, re.X)
    if pattern.search(raw):
        text = pattern.sub("Section ", raw)
        text = text.strip("., ")
        return text
    return raw


def process_units(nodes: list[dict]) -> None:
    """Recursive function to ensure that the nested units comply with the database constraints.
    Each item should be limited to 500 characters. Each caption should be limited to 500 characters.
    For items which contain a variant of the "section" label, e.g. "SEC. 1", apply uniform label,
    i.e. "Section 1".

    Args:
        nodes (list[dict]): [description]
    """
    for node in nodes:
        if node.get("item", None):
            # deal with items where the input is an integer
            # only strings can be matched in the `uniform_section_label` function
            converted = str(node["item"])

            # for SEC., # SECTION text, convert to uniform "Section"
            node["item"] = uniform_section_label(converted)

            # ensure that item text is sound, 500 as arbitrary number check
            if len(node["item"]) > 500:
                node["item"] = node["item"][:500]

        if caption := node.get("caption", None):

            # ensure that caption text is sound, 500 as arbitrary number check
            if len(node["caption"]) > 500:
                node["caption"] = caption[:500]

        if node.get("units", None):
            process_units(node["units"])


def load_statute(loc: Path) -> Optional[dict]:
    """With the passed directory, get the relevant files.

    The following files in the directory, i.e. ("pd1") are processed, if they exist

    1. `details.yaml`
    2. `extra.html`
    3. `units.yaml` (unformatted)
    4. `pd1.yaml` (preformatted Presidential Decree No. 1)

    This function combines the contents of the `details.yaml` file
    with the contents of either the `units.yaml` file or the `pd1.yaml` file.
    The resulting combination is a dictionary of key value pairs.

    Args:
        loc (Path): The source directory of the files mentioned above.

    Returns:
        Optional[dict]: The combined data found in the folder.
    """
    if not (data := set_meta(loc)):
        print(f"No details.yaml file: {loc}.")
        return None
    print(f"Details found: {loc}.")
    return set_units(loc, data)


def get_directory(
    location: Path,
    law_category: str,
    serial_num: str,
) -> Optional[Path]:
    """With the location of the files, check if a specific folder exists with the given parameters

    Args:
        location (Path): Where the source material is stored
        law_category (str): Must be either "ra", "eo", "pd", "ca", "bp", "act", "const"
        serial_num (str): A digit-like identifier, e.g. 209, 158-a, etc.

    Returns:
        Path: [description]
    """
    if law_category not in ALLOWED_CATEGORIES:
        return None

    directory = location / f"{law_category}" / f"{serial_num}"
    if not directory.exists():
        return None

    return directory


def data_from_folder(
    location: Path,
    law_category: str,
    serial_num: str,
) -> Optional[dict]:
    """If the location exists and there is data processed from such location,
    determine whether the important fields are present and then process (clean) the
    fields.

    Args:
        location (Path): Where the source material is stored
        law_category (str): Must be either "ra", "eo", "pd", "ca", "bp", "act", "const"
        serial_num (str): A digit-like identifier, e.g. 209, 158-a, etc.

    Returns:
        Optional[dict]: Cleaned data dictionary
    """

    if not (folder := get_directory(location, law_category, serial_num)):
        return None

    if not (source := load_statute(folder)):
        return None

    if source.get("units", None):
        process_units(source["units"])

    return map_to_model(law_category, source)


def map_to_model(law_category: str, source: dict) -> Optional[dict]:
    """The dictionary found in source needs to be mapped to the schema of the database.

    Args:
        law_category (str): Must be either "ra", "eo", "pd", "ca", "bp", "act", "const"
        source (dict): Pre-processed content

    Returns:
        Optional[dict]: Segregated data that matches the database schema for Statutes
    """
    for field in ["numeral", "law_title", "date", "units"]:
        if not source.get(field, None):
            print(f"Missing field: {field=}")
            return None

    # The database limit should be less than 1000 characters
    if len(source["law_title"]) > 1000:
        source["law_title"] = source["law_title"][:1000]

    if not (title_text := get_title(law_category, source["numeral"])):
        print(
            f"Could not create title text with {law_category=} and {source['numeral']=}"
        )
        return None

    return {
        "category": source["category"],
        "identifier": source["numeral"],
        "title": title_text,
        "full_title": source["law_title"],
        "specified_date": arrow.get(source["date"], "MMMM D, YYYY").date(),
        "publications": source.get("publications", None),
        "enacting_clause": source.get("enacting_clause", None),
        "whereas_clause": source.get("whereas_clause", None),
        "signers_of_law": source.get("signers_of_law", None),
        "lapse_into_law_clause": source.get("lapse_into_law_clause", None),
        "units": source["units"],
    }



import codecs
import re
from pathlib import Path

import yaml


def is_problematic(d: dict):
    if d.get("item", None) and d.get("order", None):
        match = re.search(r"\d+", d["item"])
        digit = int(match.group()) if match else 999
        return digit != d["order"]


def find_problems(loc: Path):
    target_file = loc / "units.yaml"
    if not target_file.exists():
        return False
    with open(target_file, "r") as r:
        data = {}
        data["units"] = yaml.load(r, Loader=yaml.FullLoader)
        return is_problematic(data["units"][-1])


def origin_clause_not_null(loc: Path):
    target_file = loc / "details.yaml"
    if not target_file.exists():
        return False
    with open(target_file, "r") as r:
        data = yaml.load(r, Loader=yaml.FullLoader)
        if not data["origin_clause"]:
            return False
        return (data["origin_clause"], data["origin"])


def lapse_clause_not_null(loc: Path):
    target_file = loc / "details.yaml"
    if not target_file.exists():
        return False
    with open(target_file, "r") as r:
        data = yaml.load(r, Loader=yaml.FullLoader)
        if not data["lapse_into_law_clause"]:
            return False
        return (data["lapse_into_law_clause"], data["origin"])


def first_issue_detected(loc: Path):
    with open(loc / "units.yaml", "r") as r:
        data = {}
        data["units"] = yaml.load(r, Loader=yaml.FullLoader)
        if data["units"][0] and data["units"][0].get("content", None):
            return "SEC. 2" in data["units"][0]["content"]


def order_mismatches():
    folder = Path(".") / "ra"
    locations = folder.glob("*")
    return (loc for loc in locations if find_problems(loc))


def first_item_issues():
    return [loc for loc in order_mismatches() if first_issue_detected(loc)]


def get_body(loc: Path) -> str:
    body_location = loc / "body_statute.html"
    f = codecs.open(str(body_location))
    return f.read()



import codecs
from pathlib import Path
from typing import Optional

import yaml


def set_whereas(loc: Path) -> dict:
    target = loc / "extra.html"
    if target.exists():
        f = codecs.open(str(target), "r")
        extra_content = f.read()
        title_case = "Whereas" in extra_content
        all_caps = "WHEREAS" in extra_content
        if title_case or all_caps:
            return {"whereas_clause": extra_content}
    return {"whereas_clause": None}


def set_meta(loc: Path) -> Optional[dict]:
    """The high-level config file for a statute is the `details.yaml`
    See sample `details.yaml` under the folder ../pd/1

    - numeral: '1'
    - category: pd
    - origin: [url]
    - publications: []
    - enacting_clause: 'NOW, THEREFORE, I, FERDINAND E. MARCOS, x x x'
    - signers_of_law: 'Done in the City of Manila, x x x'
    - lapse_into_law_clause: null
    - law_title: Reorganizing The Executive Branch Of The National Government
    - date: September 24, 1972
    - item: Presidential Decree No. 1

    Args:
        loc (Path): [description]

    Returns:
        Optional[dict]: [description]
    """
    try:
        details = loc / "details.yaml"
        if not details.exists():
            return None
        with open(details, "r") as r:
            return yaml.load(r, Loader=yaml.FullLoader) | set_whereas(loc)

    except FileNotFoundError:
        return None
